#!/usr/bin/env python3

#
# Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#

import argparse
from pathlib import Path
import sys

from doca_installer.utils.bf_utils import *
from doca_installer.utils.process_utils import *
from doca_installer.utils.config_processor import generate_device_cfg

PROCESS_LOCK_FILE = "/tmp/upgrade_bf.lock"
VERSION = "1.2.4"


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Upgrade BlueField image bundle"
    )
    parser.add_argument(
        "-b",
        "--image-bundle",
        help="Image bundle to install, can be bf-bundle or bf-fwbundle image.",
    )
    parser.add_argument("-r", "--rshim", help="rshim device to install on")
    parser.add_argument(
        "-c", "--config-file", help="Configuration file for all devices"
    )
    parser.add_argument(
        "-p",
        "--psid",
        help="PSID to install on. Use when wish to install on specific PSID",
    )
    parser.add_argument(
        "--exclude", 
        nargs='+', 
        help="List of RSHIM devices to exclude from installation (e.g. --exclude rshim0 rshim2 rshim3)"
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {VERSION}",
        help="script version",
    )
    parser.add_argument(
        "--show-target-fw",
        action="store_true",
        help="Extract and display target firmware versions from the image bundle and exit",
    )
    parser.add_argument(
        "--compare",
        action="store_true",
        help="Compare current device firmware versions with the provided image bundle versions and exit",
    )
    parser.add_argument(
        "--show-running-fw",
        action="store_true",
        help="Display current firmware versions of all available devices and exit",
    )
    args = parser.parse_args()

    # Check if image-bundle is required
    if not args.show_running_fw and not args.image_bundle:
        parser.error("The -b/--image-bundle argument is required unless using --show-running-fw")

    # Verify that the image-bundle file exists (only if provided)
    if args.image_bundle and not Path(args.image_bundle).exists():
        print(
            f"Error: Image bundle file '{args.image_bundle}' does not exist."
        )
        cleanup_and_exit(1)

    # Verify that the image is not a PLDM image (only if image bundle is provided)
    if args.image_bundle and is_pldm_image(args.image_bundle):
        print(
            f"Error: '{args.image_bundle}' appears to be a PLDM image format.\n"
            f"DOCA Installer does not support PLDM image formats.\n"
            f"Please use the standard bf-bundle or bf-fwbundle format instead."
        )
        cleanup_and_exit(1)

    # Verify config file
    if args.config_file and not Path(args.config_file).exists():
        print(
            f"Error: Configuration file '{args.config_file}' does not exist.\n"
        )
        cleanup_and_exit(1)
    return args


def main():
    args = parse_args()
    setup_logging("BF_Upgrade.log")
    log_dir = getattr(sys.modules['doca_installer.utils.bf_utils'], 'BF_LOG_DIR')
    print_services(VERSION)

    if args.image_bundle and args.show_running_fw:
        print("\nINFO: Both bundle image and --show-running-fw flag were provided.")
        print("      The --show-running-fw flag will be ignored.")
        print("      Tool will proceed with bundle installation and display running firmware versions.\n")
        args.show_running_fw = False

    # Handle --show-running-fw early if no bundle operations are needed
    if args.show_running_fw:
        print("\nGathering firmware versions from available devices...\n")
    else:
        # Rest of the code handles bundle-related operations
        fetched_target_bundle_fw_versions = False
        try:
            print(f"\nExtracting firmware versions from {args.image_bundle}")
            fetched_target_bundle_fw_versions = extract_firmware_versions(args.image_bundle)
            if not fetched_target_bundle_fw_versions:
                print("Warning: Failed to extract firmware versions from the image bundle - Will not be able to verify post installation versions.")
                if args.show_target_fw or args.compare:
                    print("ERROR: Cannot extract firmware versions from the image bundle.")
                    sys.exit(1)
        except Exception as e:
            print(f"ERROR: An unexpected error occurred while processing the image: {str(e)}")
            if args.show_target_fw or args.compare:
                sys.exit(1)
            else:
                print("Continuing with installation...\n")

        if fetched_target_bundle_fw_versions:
            display_target_image_firmware_versions()
            if args.show_target_fw:
                sys.exit(0)

        if not args.compare:
            print(f"\nBundle install of {args.image_bundle} image\n")
            print(f'Logs will be saved to {log_dir} directory\n')

    clear_up_procesess()

    all_bf_devices = get_bf_devices_in_parallel()
    if not all_bf_devices:
        print(
            "\nSummary:\n"
            "ERROR: Could not find any BF devices on the host via the lspci command.\n"
            "BF Device Firmware Upgrade Result: FAILED\n"
        )
        cleanup_and_exit(1)
    print('\nAvailable DPUs on host:')
    display_bf_device_basic_info(all_bf_devices)

    is_bf_devices_ok, bf_devices_upgrade_candidate = (
        check_bf_devices_prerequisites(
            all_bf_devices, specific_psid=args.psid
        )
    )

    if not is_bf_devices_ok:
        print(
            "\nSummary:\n"
            "ERROR: BF devices does not meet the prerequisites.\n"
            "BF Device Firmware Upgrade Result: FAILED\n"
        )
        cleanup_and_exit(1, all_bf_devices)

    is_bf_rshim_devices_ok, bf_devices_upgrade_candidate = (
        check_bf_rshim_devices(bf_devices_upgrade_candidate, args.rshim)
    )
    if not is_bf_rshim_devices_ok:
        print(
            "\nSummary:\n"
            "ERROR: Issue was found in RSHIM device. .\n"
            "ERROR: Please review the logs, and verify the RSHIM devices status.\n"
            "BF Devices Upgrade Result: FAILED\n"
        )
        cleanup_and_exit(1, all_bf_devices)
        
    has_devices_after_exclusion, bf_devices_upgrade_candidate = (
        filter_excluded_devices(bf_devices_upgrade_candidate, args.exclude)
    )
    if not has_devices_after_exclusion:
        cleanup_and_exit(2, all_bf_devices)

    bf_devices_upgrade_candidate = verify_device_fw_state(
        bf_devices_upgrade_candidate
    )

    # Handle display-only flags
    if args.show_running_fw:
        print("\nCurrent firmware versions:")
        display_bf_device_full_info(bf_devices_upgrade_candidate, stage="Current")
        sys.exit(0)
    elif args.compare and fetched_target_bundle_fw_versions:
        display_firmware_comparison(bf_devices_upgrade_candidate)
        sys.exit(0)

    print("\nThe following DPUs are about to be upgrade:")
    display_bf_device_full_info(bf_devices_upgrade_candidate)
    if args.config_file:
        generate_device_cfg(bf_devices_upgrade_candidate, args.config_file)

    all_devices_are_ready, ready_devices = install_bundle_orchestration(
        bf_devices_upgrade_candidate, args.image_bundle
    )
    if not ready_devices:
        print("\n\nERROR: No devices are ready after install in time")
        cleanup_and_exit(1, bf_devices_upgrade_candidate)

    if any(
        device.required_power_cycle or device.mode == "NIC" for device in bf_devices_upgrade_candidate
    ):
        print(
            "\nINFO: At least one device requires a host power cycle for new NIC Firmware to be loaded.\n"
        )

    #Post installation flow
    if ready_devices:
        if fetched_target_bundle_fw_versions:
            verify_post_installation_versions(bf_devices_upgrade_candidate)

        # Determine if BMC_REBOOT warning is needed
        bmc_reboot_warning_needed = (
            not args.config_file or
            not check_bmc_reboot_configured(ready_devices)
        )

        if bmc_reboot_warning_needed:
            print(
                "\nWARNING: BMC_REBOOT is not configured for all devices or is not set to 'yes'.\n"
                "BMC firmware updates will only take effect after a BMC reboot.\n"
            )

def entrypoint():
    lock = acquire_lock(PROCESS_LOCK_FILE)
    try:
        main()
    finally:
        release_lock(PROCESS_LOCK_FILE, lock)
