#
# Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#

import fcntl
import os
import re
import sys
from typing import List, Optional, TextIO

import psutil
from tabulate import tabulate

from doca_installer.utils.bf_utils import *


def print_services(tool_version: str) -> None:
    """
    Print various services
    """
    versions = [
        ["Tool version", tool_version],
        ["rshim", get_rshim_version()],
        ["mft", get_mst_version()],
        ["python", get_python_version()],
    ]
    header = ["Service name", "Version"]
    table = []

    for index, version in enumerate(versions, 1):
        table.append(
            [
                version[0],
                version[1],
            ]
        )

    print(tabulate(table, headers=header, tablefmt="fancy_grid"))


def cleanup_and_exit(
    exit_code: int, bf_devices: List[BFDevice] = None
) -> None:
    """
    Clean up and exit the process.

    Args:
        exit_code (int): The exit code of the process.
        bf_devices (List[BFDevice], optional): A list of BFDevice objects. Defaults to None.

    Returns:
        None
    """
    if exit_code != 0:
        print(f"ERROR: Process is terminating with exit code {exit_code}!")
        if bf_devices:
            print(
                "INFO: Running mstdump on devices, this may take few minutes"
            )
            for device in bf_devices:
                mstdump_device(device)
    sys.exit(exit_code)


def clear_up_procesess() -> None:
    """
    Terminate other running processes such as minicom or screen that may interfere
    """
    suspect_processes = ["minicom", "screen"]
    print(f"INFO: Searching for interfering processes: {', '.join(suspect_processes)}")
    current_pid = os.getpid()
    for proc in psutil.process_iter(["pid", "name", "cmdline"]):
        for suspect in suspect_processes:
            proc_cmp_name = proc.info.get("name") or ""
            if suspect in proc_cmp_name:
                if proc.info["pid"] == current_pid:
                    continue
                print(f"WARN: Found {proc.info['name']}, terminating")
                proc.terminate()
                proc.wait()


def get_python_version() -> Optional[str]:
    return sys.version.split()[0]


def get_rshim_version() -> Optional[str]:

    original_path = BFB_INSTALL_PATH
    default_path = shutil.which("bfb-install")
    if default_path and default_path != original_path:
        print(f"Warning: Found bfb-install under {default_path}. Expected path is {original_path}. Will default to using {original_path}.")

    rshim_output, rc = run_command("rshim -v")
    if rc != 0:
        return None
    rshim_version = rshim_output.split()[1]
    return rshim_version


def get_mst_version() -> Optional[str]:
    mst_output, rc = run_command("mst version")
    if rc != 0:
        return None
    mst_version = mst_output.replace(",", "").split()[2]
    return mst_version


def acquire_lock(lockfile_path: str) -> TextIO:
    """
    Aquire lock using flock for process locking.
    """
    try:
        lockfile = open(lockfile_path, "w")
        fcntl.flock(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        lockfile.write(str(os.getpid()))
        lockfile.flush()
        return lockfile
    except IOError:
        print(
            f"ERROR: Unable to acquire lock '{lockfile_path}' Another instance is running."
        )
        sys.exit(1)


def release_lock(lockfile_path: str, lockfile: TextIO):
    """
    Release the file lock.
    """
    try:
        fcntl.flock(lockfile, fcntl.LOCK_UN)
        lockfile.close()
        os.remove(lockfile_path)
    except Exception as e:
        print(f"ERROR: Failed to release lock: {e}")


def is_pldm_image(image_path: str) -> bool:
    """
    Check if the provided image is a PLDM image.
    
    A PLDM image is identified by:
    1. File extension is '.pldm'
    2. Image name contains the string 'pldm' (case-insensitive)
    
    Args:
        image_path (str): Path to the image file
        
    Returns:
        bool: True if the image is a PLDM image, False otherwise
    """
    image_name = os.path.basename(image_path).lower()
    return image_name.endswith('.pldm') or 'pldm' in image_name
